package jrummikub.control;

import java.awt.Color;

import jrummikub.model.GameSettings;
import jrummikub.model.PlayerSettings;
import jrummikub.util.Event1;
import jrummikub.util.IEvent1;
import jrummikub.util.IListener;
import jrummikub.util.IListener1;
import jrummikub.util.IListener2;
import jrummikub.view.ISettingsPanel;
import jrummikub.view.IView;

/**
 * The settings control controls the settings panel
 */
public class SettingsControl {
	private IView view;
	private Event1<GameSettings> startGameEvent = new Event1<GameSettings>();

	private GameSettings settings;

	/**
	 * Create a new settings control
	 * 
	 * @param view
	 *          the view to use
	 * @param settings
	 * 			initial game settings
	 */
	public SettingsControl(IView view, GameSettings settings) {
		this.view = view;
		this.settings = settings;
		addPlayer();
		addPlayer();
	}

	/**
	 * the start game event is emitted when the user wants to start a game and the
	 * settings made are valid
	 * 
	 * @return the event
	 */
	public IEvent1<GameSettings> getStartGameEvent() {
		return startGameEvent;
	}

	/**
	 * Start the operation of the settings control
	 */
	public void startSettings() {
		view.getSettingsPanel().setInitialMeldThreshold(settings.getInitialMeldThreshold());
		
		view.getSettingsPanel().getChangeInitialMeldThresholdEvent().add(new IListener1<Integer>() {
			@Override
			public void handle(Integer value) {
				settings.setInitialMeldThreshold(value);
				update();
			}
		});
		
		addPlayerSettingsListeners();

		view.showSettingsPanel(true);
	}

	private void addPlayerSettingsListeners() {
		view.getSettingsPanel().getStartGameEvent().add(new IListener() {
			@Override
			public void handle() {
				startGame();
			}
		});
		view.getSettingsPanel().getAddPlayerEvent().add(new IListener() {
			@Override
			public void handle() {
				addPlayer();
			}
		});
		view.getSettingsPanel().getRemovePlayerEvent()
				.add(new IListener1<Integer>() {
					@Override
					public void handle(Integer i) {
						removePlayer(i);
					}
				});
		view.getSettingsPanel().getChangePlayerColorEvent()
				.add(new IListener2<Integer, Color>() {
					@Override
					public void handle(Integer i, Color color) {
						setPlayerColor(i, color);
					}
				});
		view.getSettingsPanel().getChangePlayerNameEvent()
				.add(new IListener2<Integer, String>() {
					@Override
					public void handle(Integer i, String name) {
						setPlayerName(i, name);
					}
				});
	}

	private void addPlayer() {
		if (settings.getPlayerList().size() >= ISettingsPanel.PLAYER_COLORS.length) {
			return;
		}

		// Find unused player name
		int num = settings.getPlayerList().size();

		nameLoop: while (true) {
			num++;
			String playerName = "Spieler " + num;

			for (PlayerSettings player : settings.getPlayerList()) {
				if (playerName.equals(player.getName())) {
					continue nameLoop;
				}
			}
			break;
		}

		// Next, find unused color
		Color color = null;
		colorLoop: for (Color c : ISettingsPanel.PLAYER_COLORS) {
			color = c;
			for (PlayerSettings player : settings.getPlayerList()) {
				if (c == player.getColor()) {
					continue colorLoop;
				}
			}
			break;
		}

		settings.getPlayerList().add(new PlayerSettings("Spieler " + num, color));

		update();
	}

	private void removePlayer(int i) {
		settings.getPlayerList().remove(i);
		update();
	}

	private void setPlayerColor(int i, Color color) {
		PlayerSettings player = settings.getPlayerList().get(i);

		if (player.getColor() == color) {
			return;
		}

		for (PlayerSettings other : settings.getPlayerList()) {
			if (other.getColor() == color) {
				other.setColor(player.getColor());
				break;
			}
		}

		player.setColor(color);
		update();
	}

	private void setPlayerName(int i, String name) {
		settings.getPlayerList().get(i).setName(name);
		update();
	}

	private void update() {
		view.getSettingsPanel().enableRemovePlayerButtons(
				settings.getPlayerList().size() > 2);
		view.getSettingsPanel().enableAddPlayerButton(
				settings.getPlayerList().size() < ISettingsPanel.PLAYER_COLORS.length);

		checkSettings();

		view.getSettingsPanel().setGameSettings(settings);
	}

	private boolean checkSettings() {
		for (PlayerSettings player : settings.getPlayerList()) {
			if (player.getName().isEmpty()) {
				view.getSettingsPanel().setError(
						ISettingsPanel.SettingsError.NO_PLAYER_NAME);
				view.getSettingsPanel().enableStartGameButton(false);
				return false;
			}
		}
		for (int i = 0; i < settings.getPlayerList().size(); ++i) {
			String name = settings.getPlayerList().get(i).getName();

			for (int j = i + 1; j < settings.getPlayerList().size(); ++j) {
				if (settings.getPlayerList().get(j).getName().equals(name)) {
					view.getSettingsPanel().setError(
							ISettingsPanel.SettingsError.DUPLICATE_PLAYER_NAME);
					view.getSettingsPanel().enableStartGameButton(false);
					return false;
				}
			}
		}

		view.getSettingsPanel().setError(ISettingsPanel.SettingsError.NO_ERROR);
		view.getSettingsPanel().enableStartGameButton(true);
		return true;
	}

	private void startGame() {
		if (!checkSettings()) {
			return;
		}

		view.showSettingsPanel(false);

		startGameEvent.emit(settings);
	}
}
